module Heuristics where
  
import Board
import Data.List (transpose, foldl')

printScore :: Board -> IO ()
printScore b = do
    putStrLn $ show (smoothnessHeuristic b)
    putStrLn $ show (monotonicityHeuristic b)
    putStrLn $ show (emptyTilesHeuristic b)
    putStrLn $ show (gradientHeuristic b)
    putStrLn $ show (finishedHeuristic b)
    putStrLn $ show (stayAliveHeuristic b)
    putStrLn $ show (sumHeuristic b)
    putStrLn $ "Total: " ++ show (evalHeuristic b)
    return ()

evalHeuristic :: Board -> Float
evalHeuristic b = realToFrac $ sum $ zipWith (*) weights funcs
    where weights = [1,1,1,1,0.25,1000000,90000]
          funcs = [logBase 2 $ sumHeuristic b, 
                    smoothnessHeuristic b, 
                    monotonicityHeuristic b, 
                    logBase 2 $ gradientHeuristic b, 
                    finishedHeuristic b, 
                    stayAliveHeuristic b]

sumHeuristic :: Board -> Double
sumHeuristic b = fromIntegral $ sum $ map sum b

emptyTilesHeuristic :: Board -> Double
emptyTilesHeuristic b = fromIntegral $ length (getEmptyTiles b)

stayAliveHeuristic :: Board -> Double
stayAliveHeuristic b  
    | (getAvailableMoves b == []) = -1
    | otherwise = 0

--The following Heuristics are modified from an older project from pfp2048 2021 project
smoothnessHeuristic :: Board -> Double
smoothnessHeuristic b = (foldr smoothHelper 0 b) + (foldr smoothHelper 0 $ (transpose b)) where
  smoothHelper :: [Int] -> Double -> Double
  smoothHelper (x:y:zs) total 
    | x == y    = smoothHelper (y:zs) total + 1
    | otherwise = smoothHelper (y:zs) total
  smoothHelper _ total = total

monotonicityHeuristic :: Board -> Double
monotonicityHeuristic b = fromIntegral $ (fst $ foldl' helper  (0,0) b) + (fst $ foldl' helper (0,0) (map reverse $ transpose b)) where
    helper :: (Int,Int) -> [Int] -> (Int,Int)
    helper (total, score) (f:s:tl) 
        | tl == []              = (total, score)
        | (f == s) && (s==0)    = helper (total, score) (s:tl)
        | f >= s                = helper (total+score, score+1) (s:tl)
        | f < s                 = (total, score)

gradientHeuristic :: Board -> Double
gradientHeuristic b = realToFrac $ sum $ map sum $ zipWith (zipWith (*)) b filt 
           where filt = [[0,0,0,0], [128, 1, 0, 0], [512, 128, 1, 0], [2048, 512, 128, 1]]

finishedHeuristic :: Board -> Double
finishedHeuristic b
    | maximum (map maximum b) == 4096   = 100
    | maximum (map maximum b) == 2048   = 1
    | otherwise                         = 0